from src.greentree import *
import pytest
from io import StringIO
from contextlib import redirect_stdout


@pytest.fixture
def setup1():
    a = Node("a")
    b = Node("b")
    c = Node("c")
    d = Node("d")
    e = Node("e")
    g = Node("g")

    a.add_left(b)
    a.add_right(c)
    b.add_right(d)
    b.add_right(e)
    c.add_left("f")
    c.add_right(g)

    f = c._children[0]
    return a, b, c, d, e, f, g


def test_eq():
    a = Node("d")
    b = Node("d")
    assert a == b
    assert a == "d"
    assert "d" == a
    assert a is not b


def test_tree_prop(setup1):
    a, b, c, _, e, _, g = setup1
    g.add_right("h")
    assert a.is_root
    assert not b.is_root
    assert c.is_binary
    assert e.is_leaf
    assert g.is_single_link


def test_repr():
    a = Node("a")
    assert eval(repr(a)) == a


def test_as_dict(setup1):
    a, *_ = setup1
    assert a.as_dict == {
        "value": "a",
        "tag": None,
        "children": [
            {
                "value": "b",
                "tag": None,
                "children": [{"value": "d", "tag": None}, {"value": "e", "tag": None}],
            },
            {
                "value": "c",
                "tag": None,
                "children": [{"value": "f", "tag": None}, {"value": "g", "tag": None}],
            },
        ],
    }


def test_format(setup1):
    a, *_ = setup1
    assert (
        format(a, "pipe")
        == "└── a\n    ├── b\n    │   ├── d\n    │   └── e\n    └── c\n        ├── f\n        └── g\n"
    )

    assert (
        f"{a:dict}"
        == "{'children': [{'children': [{'tag': None, 'value': 'd'},\n                            {'tag': None, 'value': 'e'}],\n               'tag': None,\n               'value': 'b'},\n              {'children': [{'tag': None, 'value': 'f'},\n                            {'tag': None, 'value': 'g'}],\n               'tag': None,\n               'value': 'c'}],\n 'tag': None,\n 'value': 'a'}\n"
    )


def test_show(setup1):
    a, *_ = setup1
    stream = StringIO()
    with redirect_stdout(stream):
        a.show()
    assert (
        stream.getvalue().strip()
        == "└── a\n    ├── b\n    │   ├── d\n    │   └── e\n    └── c\n        ├── f\n        └── g"
    )



from src.greentree import *
import pytest


@pytest.fixture
def setup1():
    a = Node("a")
    b = Node("b")
    c = Node("c")
    d = Node("d")
    e = Node("e")
    f = Node("f")
    g = Node("g")

    a.add_right(b)
    a.add_right(c)
    b.add_right(d)
    b.add_right(e)
    c.add_right(f)
    c.add_right(g)
    return a, b, c, d, e, f, g


@pytest.fixture
def setup2():
    a = Node("a")
    b = Node("b")
    c = Node("c")
    d1 = Node("d")
    d2 = Node("d")
    a.add_right(b)
    a.add_right(c)
    b.add_right(d1)
    c.add_right(d2)
    return a, b, c, d1, d2


def test_level_order(setup1):
    a, b, c, d, e, f, g = setup1
    assert tuple(a.level_order_traversal()) == (a, b, c, d, e, f, g)


def test_pre_order(setup1):
    a, b, c, d, e, f, g = setup1
    assert tuple(a.pre_order_traversal()) == (a, b, d, e, c, f, g)


def test_post_order(setup1):
    a, b, c, d, e, f, g = setup1
    assert tuple(a.post_order_traversal()) == (d, e, b, f, g, c, a)



from collections import deque
from typing import Union, IO, Deque
from io import StringIO
from pprint import pprint


def _to_dict(node: "Node") -> dict:
    d = {"value": node.value, "tag": node.tag}
    if node._children:
        d["children"] = [_to_dict(child) for child in node._children]
    return d


def _format_tree_with_pipes(node: "Node", stream: IO, prefix="", is_last=True):
    # Connector for the current node
    connector = "└── " if is_last else "├── "
    print(f"{prefix}{connector}{node.value}", file=stream)

    # Prepare prefix for children
    new_prefix = prefix + ("    " if is_last else "│   ")

    # Iterate through children and recurse
    temp = list(reversed(node._children))
    while temp:
        child = temp.pop()
        _format_tree_with_pipes(child, stream, new_prefix, not bool(temp))


class Node:
    __slots__ = ("value", "tag", "parent", "_children")

    def __init__(self, value: object, tag: Union[str, int, None] = None):
        self.value = value
        self.tag = tag
        self.parent = None
        self._children = deque()

    @property
    def children(self) -> Deque:
        return self._children

    @property
    def is_leaf(self) -> bool:
        return not self._children

    @property
    def is_binary(self) -> bool:
        return len(self._children) == 2

    @property
    def is_root(self) -> bool:
        return self.parent is None

    @property
    def is_single_link(self) -> bool:
        return len(self._children) == 1

    @property
    def as_dict(self) -> dict:
        return _to_dict(self)

    def add_left(self, item: Union[object, "Node"]) -> None:
        if not isinstance(item, self.__class__):
            item = self.__class__(item)

        item.parent = self
        self._children.appendleft(item)

    def add_right(self, item: Union[object, "Node"]) -> None:
        if not isinstance(item, self.__class__):
            item = self.__class__(item)
        item.parent = self
        self._children.append(item)

    def level_order_traversal(self, root=None):
        if root is None:
            root = self
        queue = deque([root])
        while queue:
            node = queue.popleft()

            yield node

            if node._children:
                queue.extend(node._children)

    def pre_order_traversal(self, root=None):
        if root is None:
            root = self
        return self._dfs_traversal(root, True)

    def post_order_traversal(self, root=None):
        if root is None:
            root = self
        return self._dfs_traversal(root, False)

    def _dfs_traversal(self, root, is_preorder=True):

        if is_preorder:
            yield root

        if root.children:
            for child in root.children:
                yield from self._dfs_traversal(child, is_preorder)

        if not is_preorder:
            yield root

    def __eq__(self, other: Union[object, "Node"]):
        if issubclass(other.__class__, self.__class__):
            return self.value == other.value
        else:
            return self.value == other

    def __repr__(self):
        return f"{self.__class__.__name__}({self.value!r})"

    def __format__(self, format_spec):
        if format_spec == "pipe":
            stream = StringIO()
            _format_tree_with_pipes(self, stream)
            return stream.getvalue()
        elif format_spec == "dict":
            stream = StringIO()
            pprint(self.as_dict, stream)
            return stream.getvalue()

    def show(self) -> None:
        print(f"{self:pipe}")


__all__ = [
    "Node",
]
